#! /usr/bin/env python
"""
Created on: 05/22/2017

@author: Dorjee Gyaltsen
@brief: This script will download and install bcell-executable package.
"""
from __future__ import print_function

import os
import sys
import subprocess
from distutils.spawn import find_executable
from pip._internal.utils.misc import get_installed_distributions

NETSURFP_BIN = os.environ.get('NETSURFP_BIN', 'netsurfp')

class Configure():
    '''for configuring bcell standalone tool'''
    def __init__(self, version="3.0"):
        self.version = version
        self.project_dir = os.path.abspath(".")
        self.deps = os.path.join(self.project_dir, "deps")

    def install_requirements(self):
        '''reminder users intall requirements.'''
        requirements = os.path.join(self.project_dir, "requirements.txt")
        try:
            print("Installing requirements file (this might take some time)...")
            subprocess.check_call(["pip", "install", "-r", requirements])
        except subprocess.CalledProcessError as e:
            print(e.output)

    def check_prerequisites(self):
        prerequisites = ["tcsh", "gawk"]
        for prereq in prerequisites:
            executable = find_executable(prereq)
            if not executable:
                print("* prerequisite: '%s' is not found in your path.  Please install and then rerun the configure script" % prereq)
                sys.exit(1)
            else:
                print('prerequisite: "%s" was found.' % prereq)
        # if we get here, all prerequisites have been found
        print("All prerequisites found!")

    def check_for_bepipred2(self):
        '''check if netsurfp, 'scipy==1.2.3', and 'scikit-learn==0.17' were found '''
        executable = find_executable(NETSURFP_BIN)
        if not executable:
            print('* optional dependency: "netsurfp" was not found in your path. "netsurfp" is a dependency for method "Bepipred-2.0".'
                  ' Please install it as the direction in the "README" file before you run prediction with method "Bepipred-2.0".\n')
        else:
            print('optional dependency: "%s" was found.\n' % NETSURFP_BIN)
        missing_optional_packages_list = self.check_optional_python_modules()
        if missing_optional_packages_list:
            print('* optional python modules for the method "Bepipred-2.0" was not found: %s .' % ', '.join(missing_optional_packages_list))

    def check_python_modules(self):
        missing_packages_list = []
        installed_packages = get_installed_distributions()
        installed_packages_list = ["%s==%s" % (i.key, i.version) for i in installed_packages]
        required_packages_list = ['numpy==1.16.6', 'matplotlib==2.0.0']
        for package in required_packages_list:
            if not any(map(lambda p: p.startswith(package), installed_packages_list)):
                missing_packages_list.append(package)
        return missing_packages_list

    def check_optional_python_modules(self):
        missing_packages_list = []
        installed_packages = get_installed_distributions()
        installed_packages_list = ["%s==%s" % (i.key, i.version) for i in installed_packages]
        required_packages_list = ['scipy==1.2.3', 'scikit-learn==0.17', ]
        for package in required_packages_list:
            if not any(map(lambda p: p.startswith(package), installed_packages_list)):
                missing_packages_list.append(package)
        return missing_packages_list


    def check_python_version(self):
        if sys.version_info[0] != 3 or sys.version_info[1] < 6:
            print("* prerequisite: Python 3.6 or higher is required." )
            sys.exit(1)
        if sys.version_info[1] > 6:
            print('*  Python 3.6 is required for method "Bepipred-2.0". Please install proper Python version if you would like to run prediction with method "Bepipred-2.0".\n')
            sys.exit(1)

if __name__ == "__main__":
    config = Configure()
    config.check_prerequisites()
    config.check_python_version()
    config.check_for_bepipred2()
    
    # un-comment the line below, if you like configure script to take care of the required package install
    # config.install_requirements()
    missing_packages_list = config.check_python_modules()
    if missing_packages_list:
        print("* You must have %s package(s) installed.\nRun this command to install them:\n$ pip install %s \n" % (', '.join(missing_packages_list), ' '.join(missing_packages_list)))
    else:
        print("That's it. You're all set!")
